import torch
from torch import nn
from einops.layers.torch import Rearrange
from einops import rearrange, pack, unpack

from zeta.utils.main import default

# helper


def exists(v):
    return v is not None


def identity(t):
    return t


def divisible_by(num, den):
    return (num % den) == 0


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def is_odd(n):
    return not divisible_by(n, 2)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


class TimeUpSample2x(nn.Module):
    """
    Time Up Sample Module

    This module is used to upsample the time dimension of a tensor.

    Args:
        dim (int): The number of channels in the input tensor.
        dim_out (int): The number of channels in the output tensor.

    Returns:
        out (tensor): The upsampled tensor.

    Usage:
        >>> upsample = TimeUpSample2x(64, 128)
        >>> x = torch.randn(1, 64, 32, 32)
        >>> out = upsample(x)
        >>> out.shape
        torch.Size([1, 128, 32, 32])

    """

    def __init__(self, dim, dim_out=None):
        super().__init__()
        dim_out = default(dim_out, dim)
        conv = nn.Conv1d(dim, dim_out * 2, 1)

        self.net = nn.Sequential(
            conv, nn.SiLU(), Rearrange("b (c p) t -> b c (t p)", p=2)
        )

        self.init_conv(conv)

    def init_conv(self, conv):
        """iNIITIALIZE CONVOLUTIONAL LAYER"""
        o, i, t = conv.weight.shape
        conv_weight = torch.empty(0 // 2, i, t)
        nn.init.kaiming_normal_(conv_weight)

        conv.weight.data.copy_(conv_weight)
        nn.init.zeros_(conv.bias.data)

    def forward(self, x):
        """
        Einstein notation

        b - batch
        c - channels
        t - time
        h - height
        w - width


        """
        x = rearrange(x, "b c t h w -> b h w c t")
        x, ps = pack_one(x, "* c t")

        out = self.net(x)

        out = unpack_one(out, ps, "* c t")
        out = rearrange(out, "b h w c t -> b c t h w")
        return out



from torch import nn


class SimpleResBlock(nn.Module):
    """
    Simple residual block with GELU activation

    Args:
        channels: number of input/output channels

    Returns:
        x + proj(x) where proj is a small MLP

    Usage:
        >>> block = SimpleResBlock(256)
        >>> x = torch.randn(4, 256)
        >>> block(x).shape
        torch.Size([4, 256])

    """

    def __init__(self, channels):
        super().__init__()
        self.pre_norm = nn.LayerNorm(channels)

        self.proj = nn.Sequential(
            nn.Linear(channels, channels),
            nn.GELU(),
            nn.Linear(channels, channels),
        )

    def forward(self, x):
        """Forward pass"""
        x = self.pre_norm(x)
        return x + self.proj(x)



from torch import nn

from zeta.structs.transformer import GLU, ReluSquared


def exists(val):
    return val is not None


def default(val, default_val):
    return default_val if val is None else val


def init_zero_(layer):
    nn.init.constant_(layer.weight, 0.0)
    if exists(layer.bias):
        nn.init.constant_(layer.bias, 0.0)


class FeedForward(nn.Module):
    """
    Feedforward neural network with LayerNorms and GELU activations

    Args:
        dim (int): Input dimension
        hidden_dim (int): Hidden dimension
        dropout (float): Dropout probability

    Usage:
    >>> model = FeedForward(768, 2048, 0.1)
    >>> x = torch.randn(1, 768)
    >>> model(x).shape

    """

    def __init__(
        self,
        dim: int,
        dim_out: int = None,
        mult=4,
        glu=False,
        glu_mult_bias=False,
        swish=False,
        relu_squared=False,
        post_act_ln=False,
        dropout: float = 0.0,
        no_bias=False,
        zero_init_output=False,
    ):
        super().__init__()
        inner_dim = int(dim * mult)
        dim_out = default(dim_out, dim)

        if relu_squared:
            activation = ReluSquared()
        elif swish:
            activation = nn.SiLU()
        else:
            activation = nn.GELU()

        if glu:
            project_in = GLU(
                dim, inner_dim, activation, mult_bias=glu_mult_bias
            )
        else:
            project_in = nn.Sequential(
                nn.Linear(dim, inner_dim, bias=not no_bias), activation
            )

        if post_act_ln:
            self.ff = nn.Sequential(
                project_in,
                nn.LayerNorm(inner_dim),
                nn.Dropout(dropout),
                nn.Linear(inner_dim, dim_out, bias=not no_bias),
            )
        else:
            self.ff = nn.Sequential(
                project_in,
                nn.Dropout(dropout),
                nn.Linear(inner_dim, dim_out, bias=not no_bias),
            )

        # init last linear layer to 0
        if zero_init_output:
            init_zero_(self.ff[-1])

    def forward(self, x):
        """
        Forward pass of the feedforward network

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: Output tensor
        """
        return self.ff(x)



from zeta.nn.modules.adaptive_conv import AdaptiveConv3DMod
from zeta.nn.modules.adaptive_layernorm import AdaptiveLayerNorm
from zeta.nn.modules.cnn_text import CNNNew
from zeta.nn.modules.combined_linear import CombinedLinear
from zeta.nn.modules.convnet import ConvNet
from zeta.nn.modules.droppath import DropPath
from zeta.nn.modules.dynamic_module import DynamicModule
from zeta.nn.modules.ether import Ether
from zeta.nn.modules.exo import Exo
from zeta.nn.modules.fast_text import FastTextNew
from zeta.nn.modules.feedforward import FeedForward
from zeta.nn.modules.feedforward_network import FeedForwardNetwork
from zeta.nn.modules.flexible_mlp import CustomMLP
from zeta.nn.modules.h3 import H3Layer
from zeta.nn.modules.itca import IterativeCrossSelfAttention
from zeta.nn.modules.lang_conv_module import ConvolutionLanguageBlock
from zeta.nn.modules.layernorm import LayerNorm, l2norm
from zeta.nn.modules.leaky_relu import LeakyRELU
from zeta.nn.modules.log_ff import LogFF
from zeta.nn.modules.lora import Lora
from zeta.nn.modules.mbconv import MBConv
from zeta.nn.modules.mlp import MLP
from zeta.nn.modules.mlp_mixer import MLPMixer
from zeta.nn.modules.nebula import Nebula
from zeta.nn.modules.polymorphic_activation import PolymorphicActivation
from zeta.nn.modules.polymorphic_neuron import PolymorphicNeuronLayer
from zeta.nn.modules.prenorm import PreNorm
from zeta.nn.modules.pulsar import Pulsar
from zeta.nn.modules.residual import Residual
from zeta.nn.modules.resnet import ResNet
from zeta.nn.modules.rms_norm import RMSNorm
from zeta.nn.modules.rnn_nlp import RNNL
from zeta.nn.modules.shufflenet import ShuffleNet
from zeta.nn.modules.sig_lip import SigLipLoss
from zeta.nn.modules.simple_attention import simple_attention
from zeta.nn.modules.simple_feedforward import SimpleFeedForward
from zeta.nn.modules.simple_res_block import SimpleResBlock
from zeta.nn.modules.skipconnection import SkipConnection
from zeta.nn.modules.spacial_transformer import SpacialTransformer
from zeta.nn.modules.subln import SubLN
from zeta.nn.modules.super_resolution import SuperResolutionNet
from zeta.nn.modules.time_up_sample import TimeUpSample2x
from zeta.nn.modules.token_learner import TokenLearner
from zeta.nn.modules.unet import Unet
from zeta.nn.modules.video_autoencoder import CausalConv3d
from zeta.nn.modules.visual_expert import VisualExpert
from zeta.nn.modules.yolo import yolo
from zeta.nn.modules.swiglu import SwiGLU, SwiGLUStacked
from zeta.nn.modules.img_patch_embed import ImgPatchEmbed
from zeta.nn.modules.dense_connect import DenseBlock
from zeta.nn.modules.highway_layer import HighwayLayer
from zeta.nn.modules.multi_scale_block import MultiScaleBlock
from zeta.nn.modules.feedback_block import FeedbackBlock
from zeta.nn.modules.dual_path_block import DualPathBlock
from zeta.nn.modules.recursive_block import RecursiveBlock
from zeta.nn.modules._activations import (
    PytorchGELUTanh,
    NewGELUActivation,
    GELUActivation,
    FastGELUActivation,
    QuickGELUActivation,
    ClippedGELUActivation,
    AccurateGELUActivation,
    MishActivation,
    LinearActivation,
    LaplaceActivation,
    ReLUSquaredActivation,
)


from zeta.nn.modules.triple_skip import TripleSkipBlock
from zeta.nn.modules.dynamic_routing_block import DynamicRoutingBlock
from zeta.nn.modules.gated_residual_block import GatedResidualBlock
from zeta.nn.modules.stochastic_depth import StochasticSkipBlocK


from zeta.nn.modules.quantized_layernorm import QuantizedLN
from zeta.nn.modules.slerp_model_merger import SLERPModelMerger
from zeta.nn.modules.avg_model_merger import AverageModelMerger


# from zeta.nn.modules.img_reshape import image_reshape
# from zeta.nn.modules.flatten_features import flatten_features
# from zeta.nn.modules.scaled_sinusoidal import ScaledSinuosidalEmbedding
# from zeta.nn.modules.scale import Scale
# from zeta.nn.modules.scalenorm import ScaleNorm
# from zeta.nn.modules.simple_rmsnorm import SimpleRMSNorm
# from zeta.nn.modules.gru_gating import GRUGating
# from zeta.nn.modules.shift_tokens import ShiftTokens
# from zeta.nn.modules.swarmalator import simulate_swarmalators
# from zeta.nn.modules.transformations import image_transform
# from zeta.nn.modules.squeeze_excitation import SqueezeExcitation
# from zeta.nn.modules.clex import Clex

__all__ = [
    "CNNNew",
    "CombinedLinear",
    "ConvNet",
    "DropPath",
    "DynamicModule",
    "Exo",
    "FastTextNew",
    "FeedForwardNetwork",
    "LayerNorm",
    "l2norm",
    "Lora",
    "MBConv",
    "MLP",
    "Pulsar",
    "Residual",
    "ResNet",
    "RMSNorm",
    "RNNL",
    "ShuffleNet",
    "simple_attention",
    "SpacialTransformer",
    "SubLN",
    "SuperResolutionNet",
    "TokenLearner",
    "yolo",
    "Ether",
    "Nebula",
    "AdaptiveConv3DMod",
    "TimeUpSample2x",
    "CausalConv3d",
    "SimpleResBlock",
    "SigLipLoss",
    "SimpleFeedForward",
    "Unet",
    "VisualExpert",
    "FeedForward",
    "SkipConnection",
    "LogFF",
    "PolymorphicNeuronLayer",
    "CustomMLP",
    "PolymorphicActivation",
    "PreNorm",
    "IterativeCrossSelfAttention",
    "ConvolutionLanguageBlock",
    "H3Layer",
    "MLPMixer",
    "LeakyRELU",
    "AdaptiveLayerNorm",
    "SwiGLU",
    "SwiGLUStacked",
    "ImgPatchEmbed",
    "DenseBlock",
    "HighwayLayer",
    "MultiScaleBlock",
    "FeedbackBlock",
    "DualPathBlock",
    "RecursiveBlock",
    "PytorchGELUTanh",
    "NewGELUActivation",
    "GELUActivation",
    "FastGELUActivation",
    "QuickGELUActivation",
    "ClippedGELUActivation",
    "AccurateGELUActivation",
    "MishActivation",
    "LinearActivation",
    "LaplaceActivation",
    "ReLUSquaredActivation",
    "TripleSkipBlock",
    "DynamicRoutingBlock",
    "GatedResidualBlock",
    "StochasticSkipBlocK",
    "QuantizedLN",
    "SLERPModelMerger",
    "AverageModelMerger",
]



import torch
from torch import nn


class MultiModalFusion(nn.Module):
    def forward(self, x, y):
        return torch.einsum("bi, bj -> bij", x, y)


# # #random
# x = torch.rand(1, 3)
# y = torch.rand(1, 3)
# model = MultiModalFusion()
# out = model(x, y)
# print(out.shape)



import torch
from torch import nn
from einops.layers.torch import Rearrange
import torch.nn.functional as F


class ShuffleNet(nn.Module):
    """

    ShuffleNet implementation.


    Usage:
        from zeta.nn import ShuffleNet

        x = torch.randn(1, 3, 224, 224)
        net = ShuffleNet()
        net(x)


    """

    def __init__(
        self,
        in_channels,
        out_channels,
        groups=3,
        grouped_conv=True,
        combine="add",
    ):
        super().__init__()
        first_1x1_groups = groups if grouped_conv else 1
        bottle_neck_channels = out_channels // 4

        self.combine = combine

        if combine == "add":
            # shuffleunit
            self.left = Rearrange("...->...")  # identity
            depthwise_stride = 1

        else:
            # shuffleunit figure 2x
            self.left = nn.AvgPool2d(kernel_size=3, stride=2, padding=1)
            depthwise_stride = 2

            # Ensure output of concat has the same channels as original output channels
            out_channels -= in_channels
            assert out_channels > 0

        conv1x1 = nn.Conv2d(
            in_channels,
            bottle_neck_channels,
            kernel_size=1,
            groups=first_1x1_groups,
            bias=False,
        )

        conv3x3_depthwise = nn.Conv2d(
            bottle_neck_channels,
            bottle_neck_channels,
            kernel_size=3,
            stride=depthwise_stride,
            padding=1,
            groups=bottle_neck_channels,
            bias=False,
        )

        self.right = nn.Sequential(
            # use a 1x1 grouped or non grouped convolution to reduce input channels
            # to bottleneck channels as in a resnet bottleneck module
            conv1x1(in_channels, bottle_neck_channels, groups=first_1x1_groups),
            nn.BatchNorm2d(bottle_neck_channels),
            nn.ReLU(inplace=True),
            # channels shuffle
            Rearrange("b (c1 c2) h w -> b (c2 c1) h w", c1=groups),
            # 3x3 depthwise convolution
            conv3x3_depthwise(
                bottle_neck_channels,
                stride=depthwise_stride,
                groups=bottle_neck_channels,
            ),
            nn.BatchNorm2d(bottle_neck_channels),
            # use 1x1 grouped conbolution to expand from
            # bottleneck channels to out channels
            conv1x1(bottle_neck_channels, out_channels, groups=groups),
            nn.BatchNorm2d(out_channels),
        )

    def forward(self, x):
        """
        Run the forward pass.

        Usage:
            net = ShuffleNet()
            net(x)
        """
        if self.combine == "add":
            combined = self.left(x) + self.right(x)
        else:
            combined = torch.cat([self.left(x), self.right(x)], dim=1)
        return F.relu(combined, inplace=True)



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class MultiModalCrossAttention(nn.Module):
    """
    Multi-modal cross attention module for integrating text and image features.

    Args:
    - dim (int): Hidden dimension of the input.
    - num_heads (int): Number of heads for multi-head attention.
    - dropout_rate (float): Dropout probability.
    - normalize_qk (bool): Whether to normalize the query and key vectors.

    Usage:
    - Instantiate the module and pass text and image hidden states to it.
    """

    def __init__(
        self,
        dim,
        num_heads,
        dropout_rate=0.3,
        normalize_qk=True,
        img_size=(32, 32),
        channels=3,
    ):
        super().__init__()

        self.dim = dim
        self.head_dim = dim // num_heads
        self.normalize_qk = normalize_qk

        self.dropout = nn.Dropout(dropout_rate)
        self.norm = nn.LayerNorm(dim)

        # Projection layers for text-to-image attention
        self.query_proj = nn.Linear(dim, dim)
        self.key_proj = nn.Linear(dim, dim)
        self.value_proj = nn.Linear(dim, dim)

        # Projection layers for image-to-text attention
        self.query_proj_reverse = nn.Linear(dim, dim)
        self.key_proj_reverse = nn.Linear(dim, dim)
        self.value_proj_reverse = nn.Linear(dim, dim)

        # Output linear layer
        self.output_linear = nn.Linear(2 * dim, dim)

        # Additional layer to match the image feature dimension
        self.image_to_feature_dim = nn.Linear(
            channels * img_size[0] * img_size[1], dim
        )

    def forward(self, text_hidden, image_hidden):
        """
        text_hidden: Hidden states from text model.
        image_hidden: Hidden states from image model (4D tensor).
        """

        # Flatten image features and project to the correct dimension
        image_hidden = rearrange(image_hidden, "b c h w -> b (h w) c")
        image_hidden = self.image_to_feature_dim(image_hidden)

        # Text-to-Image Attention
        query = self.query_proj(text_hidden)
        key = self.key_proj(image_hidden)
        value = self.value_proj(image_hidden)

        if self.normalize_qk:
            query = self.norm(query)
            key = self.norm(key)

        attn_weights = F.softmax(
            torch.matmul(query, key.transpose(-2, -1)) / (self.head_dim**0.5),
            dim=-1,
        )
        attn_weights = self.dropout(attn_weights)
        text_to_image = torch.matmul(attn_weights, value)

        # Image-to-Text Attention
        query_reverse = self.query_proj_reverse(image_hidden)
        key_reverse = self.key_proj_reverse(text_hidden)
        value_reverse = self.value_proj_reverse(text_hidden)

        if self.normalize_qk:
            query_reverse = self.norm(query_reverse)
            key_reverse = self.norm(key_reverse)

        attn_weights_reverse = F.softmax(
            torch.matmul(query_reverse, key_reverse.transpose(-2, -1))
            / (self.head_dim**0.5),
            dim=-1,
        )
        attn_weights_reverse = self.dropout(attn_weights_reverse)
        image_to_text = torch.matmul(attn_weights_reverse, value_reverse)

        # Concatenate and pass through linear layer
        combined_output = torch.cat((text_to_image, image_to_text), dim=-1)
        output = self.output_linear(combined_output)

        return output

    # Parameters for demonstration


batch_size = 32
text_seq_length = 128
image_height, image_width = 32, 32
channels = 3
feature_dim = 512
num_heads = 8

# Initialize the MultiModalCrossAttention module
cross_attn = MultiModalCrossAttention(
    dim=feature_dim,
    num_heads=num_heads,
    img_size=(image_height, image_width),
    channels=channels,
)

# Generate random text features: [batch_size, text_seq_length, feature_dim]
text_features = torch.randn(batch_size, text_seq_length, feature_dim)

# Generate random image features: [batch_size, channels, image_height, image_width]
image_features = torch.randn(batch_size, channels, image_height, image_width)

# Forward pass
output = cross_attn(text_features, image_features)

# Output shape
print(
    f"Output Shape: {output.shape}"
)  # Expected shape: [batch_size, text_seq_length, feature_dim]



import torch
from torch import nn


class RecursiveBlock(nn.Module):
    def __init__(self, modules, iters, *args, **kwargs):
        """
        Initializes a RecursiveBlock module.

        Args:
            modules (nn.Module): The module to be applied recursively.
            iters (int): The number of iterations to apply the module.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__()
        self.modules = modules
        self.iters = iters

    def forward(self, x: torch.Tensor):
        """
        Forward pass of the RecursiveBlock module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor after applying the module recursively.
        """
        for _ in range(self.iters):
            x = self.modules(x)
        return x



import torch
import torch.nn as nn
import unittest

from zeta.nn.modules.dense_connect import DenseBlock


class DenseBlockTestCase(unittest.TestCase):
    def setUp(self):
        self.submodule = nn.Linear(10, 5)
        self.dense_block = DenseBlock(self.submodule)

    def test_forward(self):
        x = torch.randn(32, 10)
        output = self.dense_block(x)

        self.assertEqual(output.shape, (32, 15))  # Check output shape
        self.assertTrue(
            torch.allclose(output[:, :10], x)
        )  # Check if input is preserved
        self.assertTrue(
            torch.allclose(output[:, 10:], self.submodule(x))
        )  # Check submodule output

    def test_initialization(self):
        self.assertEqual(
            self.dense_block.submodule, self.submodule
        )  # Check submodule assignment

    def test_docstrings(self):
        self.assertIsNotNone(
            DenseBlock.__init__.__doc__
        )  # Check if __init__ has a docstring
        self.assertIsNotNone(
            DenseBlock.forward.__doc__
        )  # Check if forward has a docstring


if __name__ == "__main__":
    unittest.main()



from torch import nn
from einops import rearrange, pack, unpack

# utils
# helper


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def identity(t):
    return t


def divisible_by(num, den):
    return (num % den) == 0


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def is_odd(n):
    return not divisible_by(n, 2)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


class SpatialDownsample(nn.Module):
    """
    Spatial Downsample Module
    -------------------------

    This module is used to downsample the spatial dimension of a tensor.
    It is used in the ResNet architecture to downsample the spatial dimension
    of the input tensor by a factor of 2.

    Args:
        dim (int): The number of channels in the input tensor.
        dim_out (int): The number of channels in the output tensor.
        kernel_size (int): The size of the kernel used to downsample the input tensor.

    Returns:
        out (tensor): The downsampled tensor.

    Usage:
        >>> downsample = SpatialDownsample(64, 128, 3)
        >>> x = torch.randn(1, 64, 32, 32)
        >>> out = downsample(x)
        >>> out.shape
        torch.Size([1, 128, 16, 16])

    """

    def __init__(
        self,
        dim,
        dim_out=None,
        kernel_size=3,
    ):
        super().__init__()
        dim_out = default(dim_out, dim)
        self.conv = nn.Conv3d(
            dim,
            dim_out,
            kernel_size=kernel_size,
            stride=2,
            padding=kernel_size // 2,
        )

    def forward(self, x):
        """
        Forward pass of the SpatialDownsample module.

        """
        x = rearrange(x, "b c t h w -> b t c h w")
        x, ps = pack_one(x, "* c h w")

        out = self.conv(x)

        out = unpack_one(out, ps, "* c h w")
        out = rearrange(out, "b t c h w -> b c t h w")
        return out



from zeta.ops.unitwise_norm import unitwise_norm

from zeta.ops.softmax import (
    standard_softmax,
    # selu softmax,
    selu_softmax,
    # 2. Sparsemax,
    sparsemax,
    # 3. Local Softmax,
    local_softmax,
    # 4. Fast Softmax,
    fast_softmax,
    # 5. Sparse Softmax,
    sparse_softmax,
    # 6. gumbelmax,
    gumbelmax,
    # 7. Softmax with temp,
    temp_softmax,
    # 8. logit scaled softmax,
    logit_scaled_softmax,
    # 9. norm exponential softmax,
    norm_exp_softmax,
)

__all__ = [
    "standard_softmax",
    # selu softmax,
    "selu_softmax",
    # 2. Sparsemax,
    "sparsemax",
    # 3. Local Softmax,
    "local_softmax",
    # 4. Fast Softmax,
    "fast_softmax",
    # 5. Sparse Softmax,
    "sparse_softmax",
    # 6. gumbelmax,
    "gumbelmax",
    # 7. Softmax with temp,
    "temp_softmax",
    # 8. logit scaled softmax,
    "logit_scaled_softmax",
    # 9. norm exponential softmax,
    "norm_exp_softmax",
    "unitwise_norm",
]



from torch.optim.optimizer import Optimizer


class GradientEquilibrum(Optimizer):
    """
    Gradient Equilibrum optimizer

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate
        max_iterations (int, optional): maximum number of iterations to find equilibrium
        tol (float, optional): tolerance for equilibrium
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)

    Example:
        >>> optimizer = GradientEquilibrum(model.parameters(), lr=0.1)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    """

    def __init__(
        self,
        params,
        lr: float = 0.01,
        max_iterations: int = 1000,
        tol=1e-7,
        weight_decay=0.0,
    ):
        defaults = dict(
            lr=lr,
            max_iterations=max_iterations,
            tol=tol,
            weight_decay=weight_decay,
        )
        super(GradientEquilibrum, self).__init__(params, defaults)

    def step(self, closure=None):
        """
        Step function for Gradient Equilibrum optimizer

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        Returns:
            loss (float): loss value


        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad.data
                if group["weight_decay"] != 0:
                    grad.add(p.data, alpha=group["weight_decay"])

                # Gradient Equilibrium
                equilibrum_grad = grad - grad.mean()
                p.data -= group["lr"] * equilibrum_grad
        return loss

    def clip_grad_value(self, clip_value):
        """
        CLIp gradient value


        """
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                p.grad.data.clamp_(-clip_value, clip_value)

    def add_weight_decay(self, weight_decay):
        """
        Add weight decay to the optimizer


        """
        for group in self.param_groups:
            group["weight_decay"] = weight_decay

    def state_dict(self):
        return {
            "state": self.state,
            "param_groups": self.param_groups,
        }

    def load_state_dict(self, state_dict):
        """Loads the optimizer state."""
        self.param_groups = state_dict["param_groups"]
        self.statet = state_dict["state"]



from functools import wraps
from typing import Callable


def enforce_types(func: Callable) -> Callable:
    """
    A decorator to enforce type checks on the input parameters of a function based on its annotations.

    If a parameter doesn't have a type annotation, it can be of any type.

    Args:
        func (Callable): The function whose parameters are to be checked.

    Returns:
        Callable: The wrapped function with type checks.

    Examples:
        @enforce_types
        def add(a: int, b: int) -> int:
            return a + b

        add(1, 2)  # This is fine
        add('1', '2')  # This raises a TypeError
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        arg_names = func.__code__.co_varnames[: func.__code__.co_argcount]
        arg_types = func.__annotations__

        for name, value in list(zip(arg_names, args)) + list(kwargs.items()):
            if name in arg_types and not isinstance(value, arg_types[name]):
                raise TypeError(
                    f"Argument '{name}' is not of type"
                    f" '{arg_types[name].__name__}'"
                )

        return func(*args, **kwargs)

    return wrapper



import torch
from torch import nn


class DynamicModule(nn.Module):
    """
    A container that allows for dynamic addition, removal, and modification
    of modules

    examples
    ````
    dynamic_module = DynamicModule()
    dynamic_module.add('linear', nn.Linear(10, 10))
    dynamic_module.add('relu', nn.ReLU())
    output = dynamic_module(torch.randn(1, 10))
    dynamic_module.remove('relu')

    """

    def __init__(
        self,
        forward_method=None,
    ):
        super(DynamicModule, self).__init__()
        self.module_dict = nn.ModuleDict()
        self.forward_method = forward_method

    def add(self, name, module):
        """
        Add a module to the container

        Args:
            name (str) the name of the module
            module(nn.Module) the module to add
        """
        if isinstance(name, list):
            name = ".".join(name)
        if not isinstance(module, nn.Module):
            raise ValueError("Module must be a nn.Module")
        if name in self.module_dict:
            raise ValueError("Module name must be unique")
        self.module_dict[name] = module

    def remove(self, name):
        """
        Remove a module from the container

        Args:
            name (str) the name of the module to remove
        """
        if isinstance(name, list):
            name = ".".join(name, list)
        if name not in self.module_dict:
            raise ValueError("module name does not exist")
        del self.module_dict[name]

    def forward(self, x):
        """
        Forward pass through the modules

        Args:
            x (Tensor) the input tensor

        Returns:
            Tensor: the output tensor

        """
        if self.forward_method is not None:
            return self.forward_method(self.module_dict, x)
        for module in self.module_dict.values():
            x = module(x)
        return x

    def save_state(self, path):
        torch.save(self.state_dict(), path)

    def load_state(self, path):
        self.load_state_dict(torch.load(path))



from typing import List, Type, Union

import torch
from torch import nn


def niva(
    model: nn.Module,
    model_path: str = None,
    output_path: str = None,
    quant_type: str = "dynamic",
    quantize_layers: Union[List[Type[nn.Module]], None] = None,
    dtype: torch.dtype = torch.qint8,
    *args,
    **kwargs,
):
    """Niva: Quantize a model.

    Args:
        model (nn.Module): _description_
        model_path (str, optional): _description_. Defaults to None.
        output_path (str, optional): _description_. Defaults to None.
        quant_type (str, optional): _description_. Defaults to "dynamic".
        quantize_layers (Union[List[Type[nn.Module]], None], optional): Quantize layers. Defaults to None.
        dtype (torch.dtype, optional): _description_. Defaults to torch.qint8.

    Examples:
    >>> import torch
    >>> from zeta.quant import niva
    >>> from zeta.nn import QFTSPEmbedding
    >>> model = QFTSPEmbedding(100, 100)
    >>> niva(
    ...     model,
    ...     quant_type="static",
    ...     dtype=torch.qint8,
    ...     quantize_layers=[nn.Embedding],
    ...     model_path="model.pt",
    ...     output_path="model_quantized.pt"
    ... )

    """
    if not isinstance(model, nn.Module):
        raise TypeError("model must be a torch.nn.Module")
    if model_path is None:
        raise ValueError("model_path must be specified")
    if output_path is None:
        raise ValueError("output_path must be specified")
    if quant_type not in ["static", "dynamic"]:
        raise ValueError("quant_type must be either static or dynamic")
    if quantize_layers is not None:
        if not isinstance(quantize_layers, list):
            raise TypeError("quantize_layers must be a list")
        for layer in quantize_layers:
            if not isinstance(layer, type):
                raise TypeError("quantize_layers must be a list of types")
            if not issubclass(layer, nn.Module):
                raise TypeError(
                    "quantize_layers must be a list of types that are"
                    " subclasses of torch.nn.Module"
                )
    if not isinstance(dtype, torch.dtype):
        raise TypeError("dtype must be a torch.dtype")
    if dtype not in [torch.qint8, torch.quint8]:
        raise ValueError("dtype must be either torch.qint8 or torch.quint8")

    # Load the model
    model.load_state_dict(torch.load(model_path))

    # Ensure model is in eval model
    model.eval()

    # Apply quantization
    if quant_type == "dynamic":
        if quantize_layers is None:
            raise ValueError(
                "quantize_layers must be specified for dynamic quantization"
            )
        model = torch.quantization.quantize_dynamic(
            model, quantize_layers, dtype=dtype, *args, **kwargs
        )
    elif quant_type == "static":
        model.qconfig = torch.quantization.get_default_qconfig(dtype=dtype)
        torch.quantization.prepare(model, inplace=True)
        torch.quantization.convert(model, inplace=True)

    # Save the model
    torch.save(model.state_dict(), output_path)



from sumtree import SumTree
import torch
import random


class PrioritizedSequenceReplayBuffer:
    def __init__(
        self,
        state_size,
        action_size,
        buffer_size,
        device,
        eps=1e-5,
        alpha=0.1,
        beta=0.1,
        decay_window=5,
        decay_coff=0.4,
        pre_priority=0.7,
    ):
        """
        Initializes the PrioritizedRPS object.

        Args:
            state_size (int): The size of the state space.
            action_size (int): The size of the action space.
            buffer_size (int): The size of the replay buffer.
            device (str): The device to be used for computation.
            eps (float, optional): A small constant added to priorities to ensure non-zero probabilities. Defaults to 1e-5.
            alpha (float, optional): The exponent controlling the prioritization of experiences. Defaults to 0.1.
            beta (float, optional): The exponent controlling the importance sampling weights. Defaults to 0.1.
            decay_window (int, optional): The number of steps over which the priority decay is applied. Defaults to 5.
            decay_coff (float, optional): The coefficient controlling the rate of priority decay. Defaults to 0.4.
            pre_priority (float, optional): The initial priority value for new experiences. Defaults to 0.7.
        """
        self.tree = SumTree(data_size=buffer_size)

        # PESR params
        self.eps = eps
        self.alpha = alpha
        self.beta = beta
        self.max_priority = 1.0
        self.decay_window = decay_window
        self.decay_coff = decay_coff
        self.pre_priority = pre_priority

        # buffer params
        self.state = torch.empty(buffer_size, state_size, dtype=torch.float)
        self.action = torch.empty(buffer_size, action_size, dtype=torch.float)
        self.reward = torch.empty(buffer_size, dtype=torch.float)
        self.next_state = torch.empty(
            buffer_size, state_size, dtype=torch.float
        )
        self.done = torch.empty(buffer_size, dtype=torch.uint8)

        self.count = 0
        self.real_size = 0
        self.size = buffer_size

        # device
        self.device = device

    def add(self, transition):
        state, action, reward, next_state, done = transition

        # store transition index with maximum priority in sum tree
        self.tree.add(self.max_priority, self.count)

        # store transition in the buffer
        self.state[self.count] = torch.as_tensor(state)
        self.action[self.count] = torch.as_tensor(action)
        self.reward[self.count] = torch.as_tensor(reward)
        self.next_state[self.count] = torch.as_tensor(next_state)
        self.done[self.count] = torch.as_tensor(done)

        # update counters
        self.count = (self.count + 1) % self.size
        self.real_size = min(self.size, self.real_size + 1)

    def sample(self, batch_size):
        assert (
            self.real_size >= batch_size
        ), "buffer contains less samples than batch size"

        sample_idxs, tree_idxs = [], []
        priorities = torch.empty(batch_size, 1, dtype=torch.float)

        segment = self.tree.total_priority / batch_size
        for i in range(batch_size):
            a, b = segment * i, segment * (i + 1)

            cumsum = random.uniform(a, b)
            # sample_idx is a sample index in buffer, needed further to sample actual transitions
            # tree_idx is a index of a sample in the tree, needed further to update priorities
            tree_idx, priority, sample_idx = self.tree.get(cumsum)

            priorities[i] = priority
            tree_idxs.append(tree_idx)
            sample_idxs.append(sample_idx)
        """
        Note:
        The priorities stored in sumtree are all times alpha
        """
        probs = priorities / self.tree.total_priority
        weights = (self.real_size * probs) ** -self.beta
        weights = weights / weights.max()
        batch = (
            self.state[sample_idxs].to(self.device),
            self.action[sample_idxs].to(self.device),
            self.reward[sample_idxs].to(self.device),
            self.next_state[sample_idxs].to(self.device),
            self.done[sample_idxs].to(self.device),
        )
        return batch, weights, tree_idxs

    def update_priorities(self, data_idxs, abs_td_errors):
        """
        when we get the TD-error, we should update the transition priority p_j
        And update decay_window's transition priorities
        """
        if isinstance(abs_td_errors, torch.Tensor):
            abs_td_errors = abs_td_errors.detach().cpu().numpy()

        for data_idx, td_error in zip(data_idxs, abs_td_errors):
            # first update the batch: p_j
            # p_j <- max{|delta_j| + eps, pre_priority * p_j}
            old_priority = (
                self.pre_priority
                * self.tree.nodes[data_idx + self.tree.size - 1]
            )
            priority = (td_error + self.eps) ** self.alpha
            priority = max(priority, old_priority)
            self.tree.update(data_idx, priority)
            self.max_priority = max(self.max_priority, priority)

        # And then apply decay
        if self.count >= self.decay_window:
            # count points to the next position
            # count means the idx in the buffer and number of transition
            for i in reversed(range(self.decay_window)):
                idx = (self.count - i - 1) % self.size
                decayed_priority = priority * (self.decay_coff ** (i + 1))
                tree_idx = idx + self.tree.size - 1
                existing_priority = self.tree.nodes[tree_idx]
                self.tree.update(idx, max(decayed_priority, existing_priority))



# Import necessary libraries
import torch
from torch import nn


# Define a utility function for the masked fill to avoid overflows
def mask_fill(value, mask, fill_value):
    return value.masked_fill(mask, fill_value)


# Define the asynchronized softmax function
def asynchronized_softmax(Q, K, V, unified_max_value):
    """
    Perform the asynchronized softmax operation with a unified max value.

    :param Q: Query matrix
    :param K: Key matrix
    :param V: Value matrix
    :param unified_max_value: A scalar value to stabilize the softmax computation
    :return: Weighted attention scores after applying softmax
    """
    # Step 1: Compute attention scores by multiplying Q with the transpose of K
    attention_scores = torch.matmul(Q, K.transpose(-2, -1))

    # Step 2: Subtract unified_max_value from attention scores to avoid overflow
    attention_scores_sub_max = attention_scores - unified_max_value

    # Step 3: Asynchronously calculate the exponentials for each element
    exp_attention_scores = torch.exp(attention_scores_sub_max)

    # Step 4: Apply mask to avoid recomputation due to overflow
    attention_mask = (attention_scores_sub_max > unified_max_value) | (
        attention_scores_sub_max < -unified_max_value
    )
    exp_attention_scores = mask_fill(exp_attention_scores, attention_mask, 0.0)

    # Step 5: Compute denominators for softmax
    attention_scores_denominator = torch.sum(
        exp_attention_scores, dim=-1, keepdim=True
    )

    # Step 6: Calculate softmax asynchronously
    attention_softmax = exp_attention_scores / attention_scores_denominator

    # Step 7: Apply softmax to Value matrix
    attention_output = torch.matmul(attention_softmax, V)

    return attention_output


# Define the main class for the attention mechanism
class AsynchronizedAttention(nn.Module):
    def __init__(self, d_model, n_heads, unified_max_value):
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.unified_max_value = unified_max_value
        self.head_dim = d_model // n_heads

        # Linear layers for Q, K, V projections
        self.qkv_proj = nn.Linear(d_model, d_model * 3)

    def forward(self, x):
        batch_size, seq_length, _ = x.size()

        # Project input to Q, K, V
        qkv = self.qkv_proj(x).view(
            batch_size, seq_length, self.n_heads, 3 * self.head_dim
        )
        Q, K, V = qkv.chunk(3, dim=-1)

        # Apply the asynchronized softmax to compute attention
        attention_output = asynchronized_softmax(
            Q, K, V, self.unified_max_value
        )

        return attention_output


# Example usage
if __name__ == "__main__":
    # Define the parameters
    batch_size, seq_length, d_model, n_heads = 2, 16, 512, 8
    unified_max_value = torch.tensor(
        6.0
    )  # This value should be set based on the dataset/model

    # Create random tensors for Q, K, and V
    Q = torch.randn(batch_size, seq_length, d_model)
    K = torch.randn(batch_size, seq_length, d_model)
    V = torch.randn(batch_size, seq_length, d_model)

    # Initialize the AsynchronizedAttention module
    attention_module = AsynchronizedAttention(
        d_model, n_heads, unified_max_value
    )

    # Compute the attention output
    attention_output = attention_module(Q)
    print("Attention Output Shape:", attention_output)



from torch import nn


class ConvolutionLanguageBlock(nn.Module):
    """
    Convolutional block for language modeling.
    --------------------------------------------
    A convolutional block that consists of multiple 1D convolutional layers,
    optional batch normalization, dropout, and a flexible choice of activation functions.
    This block is designed to maintain the input's dimensionality through the network,
    making it suitable for tasks that require consistent input and output dimensions.

    Parameters:
    - in_channels (int): Number of channels in the input tensor.
    - out_channels (int): Number of channels produced by the convolution.
    - kernel_size (int): Size of the convolving kernel.
    - num_layers (int, optional): Number of convolutional layers. Default: 1
    - stride (int, optional): Stride of the convolution. Default: 1
    - padding (int, optional): Zero-padding added to both sides of the input. Default: 1
    - dilation (int, optional): Spacing between kernel elements. Default: 1
    - activation (str, optional): Type of activation function. Options: 'relu', 'gelu'. Default: 'relu'
    - use_batchnorm (bool, optional): If True, includes batch normalization. Default: False
    - dropout (float, optional): Dropout rate. Default: 0.0

    Examples:
        >>> import torch
        >>> from attnconv.main import ConvolutionLanguageBlock
        >>> x = torch.randn(1, 512, 1024)
        >>> block = ConvolutionLanguageBlock(512, 512, 3, 1, 1, 1)
        >>> out = block(x)
        >>> out.shape
        torch.Size([1, 512, 1024])
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        padding,
        depth=1,
        stride=1,
        activation="gelu",
        batchnorm=False,
        dilation=1,
        dropout=0.1,
    ):
        super(ConvolutionLanguageBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.depth = depth
        self.stride = stride
        self.activation = activation
        self.batchnorm = batchnorm
        self.dilation = dilation

        layers = []
        for _ in range(depth):
            layers.append(
                nn.Conv1d(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=stride,
                    padding=padding,
                    dilation=dilation,
                )
            )
            if batchnorm:
                layers.append(nn.BatchNorm1d(out_channels))
            if activation == "relu":
                layers.append(nn.ReLU())
            elif activation == "gelu":
                layers.append(nn.GELU())
            if dropout > 0:
                layers.append(nn.Dropout(dropout))
            in_channels = out_channels  # For stacking layers

        self.conv_layers = nn.Sequential(*layers)

    def forward(self, x):
        """Forward pass with residual connection.

        Args:
            x (_type_): _description_

        Returns:
            _type_: _description_
        """
        # Apply residual connection if dimensions match
        residual = x if x.size(1) == self.conv_layers[0].in_channels else None

        # Apply convolutional layers
        x = self.conv_layers(x)

        # Apply residual connection
        if residual is not None:
            x = x + residual

        # Return output
        return x



import torch
from torch import nn


class PositionInterpolationEmbeddings(nn.Module):
    """
    PositionInterpolation
    Overview
    ========
    Positional embeddings that interpolate between sinusoidal and learned embeddings.

    Parameters
    ==========
    dim: int
        Dimension of the input embedding.
    max_positions: int
        Maximum number of positions to embed.
    base: int
        Base of the sinusoidal embedding.
    device: torch.device
        Device to store the embeddings on.

    Attributes
    ==========
    inv_freq: torch.Tensor
        Cached inverse frequencies.
    max_seq_len_cached: int
        Maximum sequence length cached.
    scale: float
        Scale of the sinusoidal embedding.
    cos_cached: torch.Tensor
        Cached cosine values.
    sin_cached: torch.Tensor
        Cached sine values.

    Methods
    =======
    forward(x, seq_len=None)
        Forward pass of the PositionInterpolationEmbeddings.


    """

    def __init__(
        self,
        dim: int = None,
        max_positions: int = 2048,
        base: int = 10000,
        device=None,
    ):
        super().__init__()
        inv_freq = 1.0 / (
            base ** (torch.arange(0, dim, 2).float().to(device) / dim)
        )
        self.register_buffer("inv_freq", inv_freq)

        max_pos_embeds = 8192

        # build here => jit trace
        self.max_seq_len_cached = max_pos_embeds
        t = torch.arange(
            self.max_len_cached,
            device=self.inv_freq.device,
            dtype=self.inv_freq.dtype,
        )

        self.scale = 1 / 4
        t *= self.scale

        freqs = torch.einsum("i,j->ij", t, self.inv_freq)
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer(
            "cos_cached", emb.cos()[None, None, :, :], persistent=False
        )
        self.register_buffer(
            "sin_cached", emb.sin()[None, None, :, :], persistent=False
        )

    def forward(self, x, seq_len=None):
        if seq_len > self.max_seq_len_cached:
            self.max_seq_len_cached = seq_len
            t = torch.arange(
                self.max_seq_len_cached,
                device=x.device,
                dtype=self.inv_freq.dtype,
            )

            t *= self.scale
            freqs = torch.einsum("i,j->ij", t, self.inv_freq)

            emb = torch.cat((freqs, freqs), dim=-1).to(x.device)
            self.register_buffer(
                "cos_cached", emb.cos()[None, None, :, :], persistent=False
            )
            self.register_buffer(
                "sin_cached", emb.sin()[None, None, :, :], persistent=False
            )

        return (
            self.cos_cached[:, :, :seq_len, ...].to(dtype=x.dtype),
            self.sin_cached[:, :, :seq_len, ...].to(dtype=x.dtype),
        )



from typing import Optional
import torch

from torch import nn
import math


def compute_entropy_safe(
    p: torch.Tensor, minus_p: torch.Tensor
) -> torch.Tensor:
    """
    Computes the entropy of a Bernoulli distribution with probability `p`.

    Parameters
    ----------
    p : torch.Tensor
            The probability of the Bernoulli distribution. Must be in the range (0, 1).
    minus_p : torch.Tensor
            the pre-computed value of 1 - `p`. Will be, by definition, in the range (0, 1).

    Returns
    -------
    torch.Tensor
            The entropy of the Bernoulli distribution.
    """
    EPSILON = 1e-6
    p = torch.clamp(p, min=EPSILON, max=1 - EPSILON)
    minus_p = torch.clamp(minus_p, min=EPSILON, max=1 - EPSILON)

    return -p * torch.log(p) - minus_p * torch.log(minus_p)


class LogFF(nn.Module):
    """
    An implementation of fast feedforward networks from the paper "Fast Feedforward Networks".

    Args:
        input_width (int): The width of the input, i.e. the size of the last dimension of the tensor passed into `forward()`.
        leaf_width (int): The width of each leaf of this FFF.
        output_width (int): The width of the output, i.e. the size of the last dimension of the tensor returned by `forward()`.
        depth (int): The depth of the FFF tree. Will result to 2**depth leaves.
        activation (torch.nn.Module, optional): The activation function to use. Defaults to `torch.nn.ReLU()`.
        dropout (float, optional): The probability to use for the dropout at the leaves after the activations have been computed. Defaults to 0.0.
                Plays no role if self.training is False.
        train_hardened (bool, optional): Whether to use hardened decisions during training. Defaults to False.
        region_leak (float, optional): The probability of a region to leak to the next region at each node. Defaults to 0.0.
                Plays no role if self.training is False.
        usage_mode (str, optional): The mode of recording usage of the leaves and nodes of this FFF.
                Must be one of ['hard', 'soft, 'none']. Defaults to 'none'.

    Raises:
        ValueError:
            - if `input_width`, `leaf_width` or `output_width` are not positive integers
            - if `depth` is not a positive integer or 0
            - if `dropout` is not in the range [0, 1]
            - if `region_leak` is not in the range [0, 1]
            - if `usage_mode` is not one of ['hard', 'soft, 'none']

    Notes:
        - The number of leaves of the FFF will be 2**depth.
        - The number of nodes of the FFF will be 2**depth - 1.
        - The region leak of >0.5 effectively reverses the roles of the left and right child at each node.
        - Dropout and region leaks are only applied during training (i.e. model.eval() will disable them).

    Examples:
        >>> import torch
        >>> from zeta.nn.modules.log_ff import LogTimeFFF
        >>> fff = LogTimeFFF(10, 20, 30, 5)
        >>> x = torch.randn(100, 10)
        >>> y = fff(x)
        >>> y.shape
        torch.Size([100, 30])
    """

    def __init__(
        self,
        input_width: int,
        leaf_width: int,
        output_width: int,
        depth: int,
        activation=nn.ReLU(),
        dropout: float = 0.0,
        train_hardened: bool = False,
        region_leak: float = 0.0,
        usage_mode: str = "none",
    ):
        """
        Initializes a fast feedforward network (FFF).

        Parameters
        ----------
        input_width : int
                The width of the input, i.e. the size of the last dimension of the tensor passed into `forward()`.
        leaf_width : int
                The width of each leaf of this FFF.
        output_width : int
                The width of the output, i.e. the size of the last dimension of the tensor returned by `forward()`.
        depth : int
                The depth of the FFF tree. Will result to 2**depth leaves.
        activation : torch.nn.Module, optional
                The activation function to use. Defaults to `torch.nn.ReLU()`.
        dropout : float, optional
                The probability to use for the dropout at the leaves after the activations have been computed. Defaults to 0.0.
                Plays no role if self.training is False.
        train_hardened : bool, optional
                Whether to use hardened decisions during training. Defaults to False.
        region_leak : float, optional
                The probability of a region to leak to the next region at each node. Defaults to 0.0.
                Plays no role if self.training is False.
        usage_mode : str, optional
                The mode of recording usage of the leaves and nodes of this FFF.
                Must be one of ['hard', 'soft, 'none']. Defaults to 'none'.

        Raises
        ------
        ValueError
                - if `input_width`, `leaf_width` or `output_width` are not positive integers
                - if `depth` is not a positive integer or 0
                - if `dropout` is not in the range [0, 1]
                - if `region_leak` is not in the range [0, 1]
                - if `usage_mode` is not one of ['hard', 'soft, 'none']

        Notes
        -----
        - The number of leaves of the FFF will be 2**depth.
        - The number of nodes of the FFF will be 2**depth - 1.
        - The region leak of >0.5 effectively reverses the roles of the left and right child at each node.
        - Dropout and region leaks are only applied during training (i.e. model.eval() will disable them).
        """
        super().__init__()
        self.input_width = input_width
        self.leaf_width = leaf_width
        self.output_width = output_width
        self.dropout = dropout
        self.activation = activation
        self.train_hardened = train_hardened
        self.region_leak = region_leak
        self.usage_mode = usage_mode

        if (
            depth < 0
            or input_width <= 0
            or leaf_width <= 0
            or output_width <= 0
        ):
            raise ValueError(
                "input/leaf/output widths and depth must be all positive"
                " integers"
            )
        if dropout < 0 or dropout > 1:
            raise ValueError("dropout must be in the range [0, 1]")
        if region_leak < 0 or region_leak > 1:
            raise ValueError("region_leak must be in the range [0, 1]")
        if usage_mode not in ["hard", "soft", "none"]:
            raise ValueError(
                "usage_mode must be one of ['hard', 'soft', 'none']"
            )

        self.depth = nn.Parameter(
            torch.tensor(depth, dtype=torch.long), requires_grad=False
        )
        self.n_leaves = 2**depth
        self.n_nodes = 2**depth - 1

        l1_init_factor = 1.0 / math.sqrt(self.input_width)
        self.node_weights = nn.Parameter(
            torch.empty(
                (self.n_nodes, input_width), dtype=torch.float
            ).uniform_(-l1_init_factor, +l1_init_factor),
            requires_grad=True,
        )
        self.node_biases = nn.Parameter(
            torch.empty((self.n_nodes, 1), dtype=torch.float).uniform_(
                -l1_init_factor, +l1_init_factor
            ),
            requires_grad=True,
        )

        l2_init_factor = 1.0 / math.sqrt(self.leaf_width)
        self.w1s = nn.Parameter(
            torch.empty(
                (self.n_leaves, input_width, leaf_width), dtype=torch.float
            ).uniform_(-l1_init_factor, +l1_init_factor),
            requires_grad=True,
        )
        self.b1s = nn.Parameter(
            torch.empty(
                (self.n_leaves, leaf_width), dtype=torch.float
            ).uniform_(-l1_init_factor, +l1_init_factor),
            requires_grad=True,
        )
        self.w2s = nn.Parameter(
            torch.empty(
                (self.n_leaves, leaf_width, output_width), dtype=torch.float
            ).uniform_(-l2_init_factor, +l2_init_factor),
            requires_grad=True,
        )
        self.b2s = nn.Parameter(
            torch.empty(
                (self.n_leaves, output_width), dtype=torch.float
            ).uniform_(-l2_init_factor, +l2_init_factor),
            requires_grad=True,
        )
        self.leaf_dropout = nn.Dropout(dropout)

        if usage_mode != "none":
            self.node_usage = nn.Parameter(
                torch.zeros((self.n_nodes,), dtype=torch.float),
                requires_grad=False,
            )
            self.leaf_usage = nn.Parameter(
                torch.zeros((self.n_leaves,), dtype=torch.float),
                requires_grad=False,
            )

    def get_node_param_group(self) -> dict:
        """
        Returns the parameters of the nodes of this FFF, coupled with their usage tensor.

        Returns
        -------
        dict
                The parameters of the nodes of this FFF, coupled with their usage tensor.
                Will have the following keys:
                        - "params": a list containing the node parameters
                        - "usage": the node usage tensor
        """

        return {
            "params": [self.node_weights, self.node_biases],
            "usage": self.node_usage,
        }

    def get_leaf_param_group(self) -> dict:
        """
        Returns the parameters of the leaves of this FFF, coupled with their usage tensor.

        Returns
        -------
        dict
                The parameters of the leaves of this FFF, coupled with their usage tensor.
                Will have the following keys:
                        - "params": a list containing the leaf parameters
                        - "usage": the node usage tensor
        """

        return {
            "params": [self.w1s, self.b1s, self.w2s, self.b2s],
            "usage": self.leaf_usage,
        }

    def training_forward(
        self,
        x: torch.Tensor,
        return_entropies: bool = False,
        use_hard_decisions: bool = False,
    ):
        """
        Computes the forward pass of this FFF during training.

        Parameters
        ----------
        x : torch.Tensor
                The input tensor. Must have shape (..., input_width).
        return_entropies : bool, optional
                Whether to return the entropies of the decisions made at each node. Defaults to False.
                If True, the mean batch entropies for each node will be returned as a tensor of shape (n_nodes,).
        use_hard_decisions : bool, optional
                Whether to use hard decisions during the forward pass. Defaults to False.
                If True, the decisions will be rounded to the nearest integer. This will effectively make the FFF tree non-differentiable.

        Returns
        -------
        torch.Tensor
                The output tensor. Will have shape (..., output_width).
        torch.Tensor, optional
                The mean batch entropies for each node. Will be returned with shape (n_nodes,) if `return_entropies` is True.
                Will not be returned if `return_entropies` is False.

        Notes
        -----
        - The FFF tree is traversed from the root to the leaves.
                At each node, the input is multiplied by the node's weight matrix and added to the node's bias vector.
                The result is passed through a sigmoid function to obtain a probability.
                The probability is used to modify the mixture of the current batch of inputs.
                The modified mixture is passed to the next node.
                Finally, the outputs of all leaves are mixed together to obtain the final output.
        - If `use_hard_decisions` is True and `return_entropies` is True, the entropies will be computed before the decisions are rounded.
        - If self.training is False, region leaks and dropout will not be applied in this function.
        - Node usage, when tracked, is computed after node leaks have been applied (but is of course also applied when there is no node leaks).

        Raises
        ------
        ValueError
                - if `x` does not have shape (..., input_width)

        See Also
        --------
        `eval_forward()`

        """
        # x has shape (batch_size, input_width)
        original_shape = x.shape
        x = x.view(-1, x.shape[-1])
        batch_size = x.shape[0]

        if x.shape[-1] != self.input_width:
            raise ValueError(
                f"input tensor must have shape (..., {self.input_width})"
            )

        hard_decisions = use_hard_decisions or self.train_hardened
        current_mixture = torch.ones(
            (batch_size, self.n_leaves), dtype=torch.float, device=x.device
        )
        entropies = (
            None
            if not return_entropies
            else torch.zeros(
                (batch_size, self.n_nodes), dtype=torch.float, device=x.device
            )
        )

        if self.usage_mode != "none" and self.depth.item() > 0:
            self.node_usage[0] += batch_size

        for current_depth in range(self.depth.item()):
            platform = torch.tensor(
                2**current_depth - 1, dtype=torch.long, device=x.device
            )
            next_platform = torch.tensor(
                2 ** (current_depth + 1) - 1, dtype=torch.long, device=x.device
            )

            n_nodes = 2**current_depth
            current_weights = self.node_weights[
                platform:next_platform
            ]  # (n_nodes, input_width)
            current_biases = self.node_biases[
                platform:next_platform
            ]  # (n_nodes, 1)

            boundary_plane_coeff_scores = torch.matmul(
                x, current_weights.transpose(0, 1)
            )  # (batch_size, n_nodes)
            boundary_plane_logits = (
                boundary_plane_coeff_scores + current_biases.transpose(0, 1)
            )  # (batch_size, n_nodes)
            boundary_effect = torch.sigmoid(
                boundary_plane_logits
            )  # (batch_size, n_nodes)

            if self.region_leak > 0.0 and self.training:
                transpositions = torch.empty_like(boundary_effect).uniform_(
                    0, 1
                )  # (batch_size, n_cuts)
                transpositions = (
                    transpositions < self.region_leak
                )  # (batch_size, n_cuts)
                boundary_effect = torch.abs(
                    transpositions.float() - boundary_effect
                )  # (batch_size, n_cuts)

            not_boundary_effect = 1 - boundary_effect  # (batch_size, n_nodes)

            if return_entropies:
                platform_entropies = compute_entropy_safe(
                    boundary_effect, not_boundary_effect
                )  # (batch_size, n_nodes)
                entropies[:, platform:next_platform] = (
                    platform_entropies  # (batch_size, n_nodes)
                )

            if hard_decisions:
                boundary_effect = torch.round(
                    boundary_effect
                )  # (batch_size, n_nodes)
                not_boundary_effect = (
                    1 - boundary_effect
                )  # (batch_size, n_nodes)

            mixture_modifier = (
                torch.cat(  # this cat-fu is to interleavingly combine the two tensors
                    (
                        not_boundary_effect.unsqueeze(-1),
                        boundary_effect.unsqueeze(-1),
                    ),
                    dim=-1,
                )
                .flatten(start_dim=-2, end_dim=-1)
                .unsqueeze(-1)
            )  # (batch_size, n_nodes*2, 1)
            current_mixture = current_mixture.view(
                batch_size, 2 * n_nodes, self.n_leaves // (2 * n_nodes)
            )  # (batch_size, 2*n_nodes, self.n_leaves // (2*n_nodes))
            current_mixture.mul_(
                mixture_modifier
            )  # (batch_size, 2*n_nodes, self.n_leaves // (2*n_nodes))
            current_mixture = current_mixture.flatten(
                start_dim=1, end_dim=2
            )  # (batch_size, self.n_leaves)

            if (
                self.usage_mode != "none"
                and current_depth != self.depth.item() - 1
            ):
                if self.usage_mode == "soft":
                    current_node_usage = mixture_modifier.squeeze(-1).sum(
                        dim=0
                    )  # (n_nodes*2,)
                elif self.usage_mode == "hard":
                    current_node_usage = (
                        torch.round(mixture_modifier).squeeze(-1).sum(dim=0)
                    )  # (n_nodes*2,)
                self.node_usage[
                    next_platform : next_platform + n_nodes * 2
                ] += current_node_usage.detach()  # (n_nodes*2,)

            del (
                mixture_modifier,
                boundary_effect,
                not_boundary_effect,
                boundary_plane_logits,
                boundary_plane_coeff_scores,
                current_weights,
                current_biases,
            )

        if self.usage_mode != "none":
            if self.usage_mode == "hard":
                current_leaf_usage = torch.round(current_mixture).sum(
                    dim=0
                )  # (n_leaves,)
            else:
                current_leaf_usage = current_mixture.sum(dim=0)  # (n_leaves,)
            self.leaf_usage.data += current_leaf_usage.detach()

        element_logits = torch.matmul(
            x, self.w1s.transpose(0, 1).flatten(1, 2)
        )  # (batch_size, self.n_leaves * self.leaf_width)
        element_logits = element_logits.view(
            batch_size, self.n_leaves, self.leaf_width
        )  # (batch_size, self.n_leaves, self.leaf_width)
        element_logits += self.b1s.view(
            1, *self.b1s.shape
        )  # (batch_size, self.n_leaves, self.leaf_width)
        element_activations = self.activation(
            element_logits
        )  # (batch_size, self.n_leaves, self.leaf_width)
        element_activations = self.leaf_dropout(
            element_activations
        )  # (batch_size, self.n_leaves, self.leaf_width)
        new_logits = torch.empty(
            (batch_size, self.n_leaves, self.output_width),
            dtype=torch.float,
            device=x.device,
        )
        for i in range(self.n_leaves):
            new_logits[:, i] = (
                torch.matmul(element_activations[:, i], self.w2s[i])
                + self.b2s[i]
            )
        # new_logits has shape (batch_size, self.n_leaves, self.output_width)

        new_logits *= current_mixture.unsqueeze(
            -1
        )  # (batch_size, self.n_leaves, self.output_width)
        final_logits = new_logits.sum(dim=1)  # (batch_size, self.output_width)

        final_logits = final_logits.view(
            *original_shape[:-1], self.output_width
        )  # (..., self.output_width)

        if not return_entropies:
            return final_logits
        else:
            return final_logits, entropies.mean(dim=0)

    def forward(
        self,
        x: torch.Tensor,
        return_entropies: bool = False,
        use_hard_decisions: Optional[bool] = None,
    ):
        """
        Computes the forward pass of this FFF.
        If `self.training` is True, `training_forward()` will be called, otherwise `eval_forward()` will be called.

        Parameters
        ----------
        x : torch.Tensor
                The input tensor. Must have shape (..., input_width).
        return_entropies : bool, optional
                Whether to return the entropies of the decisions made at each node. Defaults to False.
                If True, the mean batch entropies for each node will be returned as a tensor of shape (n_nodes,).
        use_hard_decisions : bool, optional
                Whether to use hard decisions during the forward pass. Defaults to None.
                If None and `self.training` is True, will effectively be False.
                If None and `self.training` is False, will effectively be True.
                Cannot be set to False if `self.training` is False.


        Returns
        -------
        torch.Tensor
                The output tensor. Will have shape (..., output_width).
        torch.Tensor, optional
                The mean batch entropies for each node. Will be returned with shape (n_nodes,) if `return_entropies` is True.
                Will not be returned if `return_entropies` is False.

        Raises
        ------
        ValueError
                - if `x` does not have shape (..., input_width)
                - if `return_entropies` is True and `self.training` is False
                - if `use_hard_decisions` is False and `self.training` is False

        See Also
        --------
        `training_forward()`
        `eval_forward()`
        """

        if self.training:
            return self.training_forward(
                x,
                return_entropies=return_entropies,
                use_hard_decisions=(
                    use_hard_decisions
                    if use_hard_decisions is not None
                    else False
                ),
            )
        else:
            if return_entropies:
                raise ValueError("Cannot return entropies during evaluation.")
            if use_hard_decisions is not None and not use_hard_decisions:
                raise ValueError("Cannot use soft decisions during evaluation.")
            return self.eval_forward(x)

    def eval_forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Computes the forward pass of this FFF during evaluation (i.e. making hard decisions at each node and traversing the FFF in logarithmic time).

        Parameters
        ----------
        x : torch.Tensor
                The input tensor. Must have shape (..., input_width).

        Returns
        -------
        torch.Tensor
                The output tensor. Will have shape (..., output_width).

        Notes
        -----
        - Dropout and region leaks are not engaged by this method.

        """
        original_shape = x.shape
        x = x.view(-1, x.shape[-1])
        batch_size = x.shape[0]
        # x has shape (batch_size, input_width)

        current_nodes = torch.zeros(
            (batch_size,), dtype=torch.long, device=x.device
        )
        for i in range(self.depth.item()):
            plane_coeffs = self.node_weights.index_select(
                dim=0, index=current_nodes
            )  # (batch_size, input_width)
            plane_offsets = self.node_biases.index_select(
                dim=0, index=current_nodes
            )  # (batch_size, 1)
            plane_coeff_score = torch.bmm(
                x.unsqueeze(1), plane_coeffs.unsqueeze(-1)
            )  # (batch_size, 1, 1)
            plane_score = (
                plane_coeff_score.squeeze(-1) + plane_offsets
            )  # (batch_size, 1)
            plane_choices = (
                plane_score.squeeze(-1) >= 0
            ).long()  # (batch_size,)

            platform = torch.tensor(
                2**i - 1, dtype=torch.long, device=x.device
            )  # (batch_size,)
            next_platform = torch.tensor(
                2 ** (i + 1) - 1, dtype=torch.long, device=x.device
            )  # (batch_size,)
            current_nodes = (
                (current_nodes - platform) * 2 + plane_choices + next_platform
            )  # (batch_size,)

        leaves = current_nodes - next_platform  # (batch_size,)
        new_logits = torch.empty(
            (batch_size, self.output_width), dtype=torch.float, device=x.device
        )
        for i in range(leaves.shape[0]):
            leaf_index = leaves[i]
            logits = torch.matmul(
                x[i].unsqueeze(0),  # (1, self.input_width)
                self.w1s[leaf_index],  # (self.input_width, self.leaf_width)
            )  # (1, self.leaf_width)
            logits += self.b1s[leaf_index].unsqueeze(-2)  # (1, self.leaf_width)
            activations = self.activation(logits)  # (1, self.leaf_width)
            new_logits[i] = torch.matmul(
                activations, self.w2s[leaf_index]
            ).squeeze(
                -2
            )  # (1, self.output_width)

        return new_logits.view(
            *original_shape[:-1], self.output_width
        )  # (..., self.output_width)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeta',
 'zeta.cli',
 'zeta.cloud',
 'zeta.models',
 'zeta.nn',
 'zeta.nn.attention',
 'zeta.nn.biases',
 'zeta.nn.embeddings',
 'zeta.nn.modules',
 'zeta.nn.modules.xmoe',
 'zeta.ops',
 'zeta.optim',
 'zeta.quant',
 'zeta.rl',
 'zeta.structs',
 'zeta.tokenizers',
 'zeta.training',
 'zeta.utils']

package_data = \
{'': ['*']}

install_requires = \
['accelerate==0.25.0',
 'argparse>=1.4.0,<2.0.0',
 'beartype==0.16.4',
 'bitsandbytes==0.41.3.post2',
 'colt5-attention==0.10.19',
 'datasets==2.10.1',
 'einops-exts==0.0.4',
 'einops==0.7.0',
 'lion-pytorch==0.0.7',
 'pytest==7.4.2',
 'rich==13.7.0',
 'scipy==1.9.3',
 'sentencepiece==0.1.99',
 'skypilot==0.4.1',
 'tiktoken==0.5.2',
 'timm==0.6.13',
 'tokenmonster==1.1.12',
 'torch==2.1.2',
 'torchdiffeq==0.2.3',
 'torchvision',
 'tqdm==4.66.1',
 'transformers==4.36.0',
 'typing==3.7.4.3',
 'vector-quantize-pytorch==1.12.0']

entry_points = \
{'console_scripts': ['zeta = zeta.cli.main:main']}

setup_kwargs = {
    'name': 'zetascale',
    'version': '1.3.4',
    'description': 'Transformers at zeta scales',
    'long_description': '[![Multi-Modality](images/agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n![Zeta banner](images/zeta.png)\nBuild SOTA AI Models 80% faster with modular, high-performance, and scalable building blocks!\n\n[![Docs](https://readthedocs.org/projects/zeta/badge/)](https://zeta.readthedocs.io)\n\n<p>\n  <a href="https://github.com/kyegomez/zeta/blob/main/LICENSE"><img alt="MIT License" src="https://img.shields.io/badge/license-MIT-blue.svg" /></a>\n  <a href="https://pypi.org/project/zetascale"><img alt="MIT License" src="https://badge.fury.io/py/zetascale.svg" /></a>\n</p>\n\n[![GitHub issues](https://img.shields.io/github/issues/kyegomez/zeta)](https://github.com/kyegomez/zeta/issues) [![GitHub forks](https://img.shields.io/github/forks/kyegomez/zeta)](https://github.com/kyegomez/zeta/network) [![GitHub stars](https://img.shields.io/github/stars/kyegomez/zeta)](https://github.com/kyegomez/zeta/stargazers) [![GitHub license](https://img.shields.io/github/license/kyegomez/zeta)](https://github.com/kyegomez/zeta/blob/main/LICENSE)[![GitHub star chart](https://img.shields.io/github/stars/kyegomez/zeta?style=social)](https://star-history.com/#kyegomez/zeta)[![Dependency Status](https://img.shields.io/librariesio/github/kyegomez/zeta)](https://libraries.io/github/kyegomez/zeta) [![Downloads](https://static.pepy.tech/badge/zeta/month)](https://pepy.tech/project/zeta)\n\n[![Join the Agora discord](https://img.shields.io/discord/1110910277110743103?label=Discord&logo=discord&logoColor=white&style=plastic&color=d7b023)![Share on Twitter](https://img.shields.io/twitter/url/https/twitter.com/cloudposse.svg?style=social&label=Share%20%40kyegomez/zeta)](https://twitter.com/intent/tweet?text=Check%20out%20this%20amazing%20AI%20project:%20&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta) [![Share on Facebook](https://img.shields.io/badge/Share-%20facebook-blue)](https://www.facebook.com/sharer/sharer.php?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta) [![Share on LinkedIn](https://img.shields.io/badge/Share-%20linkedin-blue)](https://www.linkedin.com/shareArticle?mini=true&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&title=&summary=&source=)\n\n[![Share on Reddit](https://img.shields.io/badge/-Share%20on%20Reddit-orange)](https://www.reddit.com/submit?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&title=zeta%20-%20the%20future%20of%20AI) [![Share on Hacker News](https://img.shields.io/badge/-Share%20on%20Hacker%20News-orange)](https://news.ycombinator.com/submitlink?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&t=zeta%20-%20the%20future%20of%20AI) [![Share on Pinterest](https://img.shields.io/badge/-Share%20on%20Pinterest-red)](https://pinterest.com/pin/create/button/?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&media=https%3A%2F%2Fexample.com%2Fimage.jpg&description=zeta%20-%20the%20future%20of%20AI) [![Share on WhatsApp](https://img.shields.io/badge/-Share%20on%20WhatsApp-green)](https://api.whatsapp.com/send?text=Check%20out%20zeta%20-%20the%20future%20of%20AI%20%23zeta%20%23AI%0A%0Ahttps%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta)\n\n\n# Install\n\n`pip install zetascale`\n\n# Usage\n\n## Starting Your Journey\n\nCreating a model empowered with the aforementioned breakthrough research features is a breeze. Here\'s how to quickly materialize the renowned Flash Attention\n\n```python\nimport torch\nfrom zeta.nn import FlashAttention\n\nq = torch.randn(2, 4, 6, 8)\nk = torch.randn(2, 4, 10, 8)\nv = torch.randn(2, 4, 10, 8)\n\nattention = FlashAttention(causal=False, dropout=0.1, flash=True)\noutput = attention(q, k, v)\n\nprint(output.shape) \n\n```\n\n\n\n### `SwiGLU`\n- Powers Transformer models\n```python\nfrom zeta.nn import SwiGLUStacked\nimport torch\n\nx = torch.randn(5, 10)\nswiglu = SwiGLUStacked(10, 20)\nswiglu(x).shape\n\n```\n\n### ```RelativePositionBias```\n- ```RelativePositionBias``` quantizes the distance between two positions into a certain number of buckets and then uses an embedding to get the relative position bias. This mechanism aids in the attention mechanism by providing biases based on relative positions between the query and key, rather than relying solely on their absolute positions.\n```python\nfrom zeta.nn import RelativePositionBias\nimport torch\n\n# Initialize the RelativePositionBias module\nrel_pos_bias = RelativePositionBias()\n\n# Example 1: Compute bias for a single batch\nbias_matrix = rel_pos_bias(1, 10, 10)\n\n# Example 2: Utilize in conjunction with an attention mechanism\n# NOTE: This is a mock example, and may not represent an actual attention mechanism\'s complete implementation.\nclass MockAttention(nn.Module):\n    def __init__(self):\n        super().__init__()\n        self.rel_pos_bias = RelativePositionBias()\n\n    def forward(self, queries, keys):\n        bias = self.rel_pos_bias(queries.size(0), queries.size(1), keys.size(1))\n        # Further computations with bias in the attention mechanism...\n        return None  # Placeholder\n\n# Example 3: Modify default configurations\ncustom_rel_pos_bias = RelativePositionBias(bidirectional=False, num_buckets=64, max_distance=256, n_heads=8)\n\n```\n\n### `FeedForward`\nThe FeedForward module performs a feedforward operation on the input tensor x. It consists of a multi-layer perceptron (MLP) with an optional activation function and LayerNorm. \n\n```python\nfrom zeta.nn import FeedForward\n\nmodel = FeedForward(\n  256, \n  512, \n  glu=True, \n  post_act_ln=True, \n  dropout=0.2\n)\n\nx = torch.randn(1, 256)\n\noutput = model(x)\nprint(output.shape)\n```\n\n### `BitLinear`\n- The BitLinear module performs linear transformation on the input data, followed by quantization and dequantization. The quantization process is performed using the absmax_quantize function, which quantizes the input tensor based on the absolute maximum value, [from the paper](https://arxiv.org/abs/2310.11453)\n```python\nimport torch\nfrom torch import nn\nimport zeta.quant as qt\n\nclass MyModel(nn.Module):\n    def __init__(self):\n        super(MyModel, self).__init__()\n        self.linear = qt.BitLinear(10, 20)\n\n    def forward(self, x):\n        return self.linear(x)\n\n# Initialize the model\nmodel = MyModel()\n\n# Create a random tensor of size (128, 10)\ninput = torch.randn(128, 10)\n\n# Perform the forward pass\noutput = model(input)\n\n# Print the size of the output\nprint(output.size())  # torch.Size([128, 20])\n\n```\n\n### `PalmE`\n- This is an implementation of the multi-modal Palm-E model using a decoder llm as the backbone with an VIT image encoder to process vision, it\'s very similiar to GPT4, Kosmos, RTX2, and many other multi-modality model architectures\n\n```python\nimport torch\nfrom zeta.structs import (\n  AutoregressiveWrapper,\n  Decoder,\n  Encoder,\n  Transformer,\n  ViTransformerWrapper,\n)\n\n\nclass PalmE(torch.nn.Module):\n    """\n    PalmE is a transformer architecture that uses a ViT encoder and a transformer decoder.\n\n    Args:\n\n        image_size (int): Size of the image.\n        patch_size (int): Size of the patch.\n        encoder_dim (int): Dimension of the encoder.\n        encoder_depth (int): Depth of the encoder.\n        encoder_heads (int): Number of heads in the encoder.\n        num_tokens (int): Number of tokens.\n        max_seq_len (int): Maximum sequence length.\n        decoder_dim (int): Dimension of the decoder.\n        decoder_depth (int): Depth of the decoder.\n        decoder_heads (int): Number of heads in the decoder.\n        alibi_num_heads (int): Number of heads in the alibi attention.\n        attn_kv_heads (int): Number of heads in the attention key-value projection.\n        use_abs_pos_emb (bool): Whether to use absolute positional embeddings.\n        cross_attend (bool): Whether to cross attend in the decoder.\n        alibi_pos_bias (bool): Whether to use positional bias in the alibi attention.\n        rotary_xpos (bool): Whether to use rotary positional embeddings.\n        attn_flash (bool): Whether to use attention flash.\n        qk_norm (bool): Whether to normalize the query and key in the attention layer.\n\n    Returns:\n\n            torch.Tensor: The output of the model.\n\n    Usage:\n\nimg = torch.randn(1, 3, 256, 256)\ntext = torch.randint(0, 20000, (1, 1024))\nmodel = PalmE()\noutput = model(img, text)\nprint(output)\n\n    """\n\n    def __init__(\n        self,\n        image_size=256,\n        patch_size=32,\n        encoder_dim=512,\n        encoder_depth=6,\n        encoder_heads=8,\n        num_tokens=20000,\n        max_seq_len=1024,\n        decoder_dim=512,\n        decoder_depth=6,\n        decoder_heads=8,\n        alibi_num_heads=4,\n        attn_kv_heads=2,\n        use_abs_pos_emb=False,\n        cross_attend=True,\n        alibi_pos_bias=True,\n        rotary_xpos=True,\n        attn_flash=True,\n        qk_norm=True,\n    ):\n        super(PalmE, self).__init__()\n\n        # vit architecture\n        self.encoder = ViTransformerWrapper(\n            image_size=image_size,\n            patch_size=patch_size,\n            attn_layers=Encoder(\n                dim=encoder_dim, depth=encoder_depth, heads=encoder_heads\n            ),\n        )\n\n        # palm model architecture\n        self.decoder = Transformer(\n            num_tokens=num_tokens,\n            max_seq_len=max_seq_len,\n            use_abs_pos_emb=use_abs_pos_emb,\n            attn_layers=Decoder(\n                dim=decoder_dim,\n                depth=decoder_depth,\n                heads=decoder_heads,\n                cross_attend=cross_attend,\n                alibi_pos_bias=alibi_pos_bias,\n                alibi_num_heads=alibi_num_heads,\n                rotary_xpos=rotary_xpos,\n                attn_kv_heads=attn_kv_heads,\n                attn_flash=attn_flash,\n                qk_norm=qk_norm,\n            ),\n        )\n\n        # autoregressive wrapper to enable generation of tokens\n        self.decoder = AutoregressiveWrapper(self.decoder)\n\n    def forward(self, img: torch.Tensor, text: torch.Tensor):\n        """Forward pass of the model."""\n        try:\n            encoded = self.encoder(img, return_embeddings=True)\n            return self.decoder(text, context=encoded)\n        except Exception as error:\n            print(f"Failed in forward method: {error}")\n            raise\n\n# Usage with random inputs\nimg = torch.randn(1, 3, 256, 256)\ntext = torch.randint(0, 20000, (1, 1024))\n\n# Initiliaze the model\nmodel = PalmE()\noutput = model(img, text)\nprint(output)\n\n\n```\n\n\n### `Unet`\nUnet is a famous convolutional neural network architecture originally used for biomedical image segmentation but soon became the backbone of the generative AI Mega-revolution. The architecture comprises two primary pathways: downsampling and upsampling, followed by an output convolution. Due to its U-shape, the architecture is named U-Net. Its symmetric architecture ensures that the context (from downsampling) and the localization (from upsampling) are captured effectively.\n\n```python\nimport torch\nfrom zeta.nn import Unet  \n\n# Initialize the U-Net model\nmodel = Unet(n_channels=1, n_classes=2)\n\n# Random input tensor with dimensions [batch_size, channels, height, width]\nx = torch.randn(1, 1, 572, 572)\n\n# Forward pass through the model\ny = model(x)\n\n# Output\nprint(f"Input shape: {x.shape}")\nprint(f"Output shape: {y.shape}")\n\n\n```\n\n\n### `VisionEmbeddings`\nThe VisionEmbedding class is designed for converting images into patch embeddings, making them suitable for processing by transformer-based models. This class plays a crucial role in various computer vision tasks and enables the integration of vision data into transformer architectures!\n\n```python\nfrom zeta.nn import VisionEmbedding\nimport torch\n\n# Create an instance of VisionEmbedding\nvision_embedding = VisionEmbedding(\n  img_size=224,\n  patch_size=16,\n  in_chans=3,\n  embed_dim=768,\n  contain_mask_token=True,\n  prepend_cls_token=True,\n)\n\n# Load an example image (3 channels, 224x224)\ninput_image = torch.rand(1, 3, 224, 224)\n\n# Perform image-to-patch embedding\noutput = vision_embedding(input_image)\n\n# The output now contains patch embeddings, ready for input to a transformer model\n```\n\n\n### `niva`\n- Niva focuses on weights of certain layers (specified by quantize_layers). Ideal for models where runtime activation is variable. 👁️ Example Layers: nn.Embedding, nn.LSTM. \n\n```python\nimport torch\nfrom zeta import niva\n\n# Load a pre-trained model\nmodel = YourModelClass()\n\n# Quantize the model dynamically, specifying layers to quantize\nniva(\n    model=model,\n    model_path="path_to_pretrained_model_weights.pt",\n    output_path="quantized_model.pt",\n    quant_type="dynamic",\n    quantize_layers=[nn.Linear, nn.Conv2d],\n    dtype=torch.qint8\n)\n\n```\n\n\n### `FusedDenseGELUDense`\n- Increase model speed by 2x with this module that fuses together 2 hyper-optimized dense ops from bits and bytes and a gelu together!\n\n```python\nimport torch\nfrom zeta.nn import FusedDenseGELUDense\n\nx = torch.randn(1, 512)\nmodel = FusedDenseGELUDense(512, 1024)\nout = model(x)\nout.shape\n\n```\n\n\n### `FusedDropoutLayerNorm`\n- FusedDropoutLayerNorm is a fused kernel of dropout and layernorm to speed up FFNs or MLPS by 2X\n\n```python\nimport torch\nfrom torch import nn\nfrom zeta.nn import FusedDropoutLayerNorm\n\n# Initialize the module\nmodel = FusedDropoutLayerNorm(dim=512)\n\n# Create a sample input tensor\nx = torch.randn(1, 512)\n\n# Forward pass\noutput = model(x)\n\n# Check output shape\nprint(output.shape)  # Expected: torch.Size([1, 512])\n\n```\n\n\n### ZetaCloud\nTrain or finetune any model on any cluster in 1 click with zetacloud, just pass in your file and the GPU type and quantity you want! To gain access first `pip install zetascale` then run `zeta -h` in the terminal. [Here is the docs for more](https://zeta.apac.ai/en/latest/zeta/cloud/main/)\n\n- Flexible Pricing with pooling from many clouds\n- Easy Deployment with 1 click\n- Various options for cloud providers!\n\n```bash\nZetacloud CLI\n\noptions:\n  -h, --help            show this help message and exit\n  -t TASK_NAME, --task_name TASK_NAME\n                        Task name\n  -c CLUSTER_NAME, --cluster_name CLUSTER_NAME\n                        Cluster name\n  -cl CLOUD, --cloud CLOUD\n                        Cloud provider\n  -g GPUS, --gpus GPUS  GPUs\n  -f FILENAME, --filename FILENAME\n                        Filename\n  -s, --stop            Stop flag\n  -d, --down            Down flag\n  -sr, --status_report  Status report flag\n\n```\n\n- A simple run example code would be like:\n\n```bash\nzeta -f train.py -g A100:8\n```\n\n# Documentation\n[Click here for the documentation, it\'s at zeta.apac.ai](https://zeta.apac.ai)\n\n# 🤝 Schedule a 1-on-1 Session\nBook a [1-on-1 Session with Kye](https://calendly.com/apacai/agora), the Creator, to discuss any issues, provide feedback, or explore how we can improve Zeta for you.\n\n## Contributing\n- We need you to help us build the most re-useable, reliable, and high performance ML framework ever.\n\n- [Check out the project board here!](https://github.com/users/kyegomez/projects/7/views/2)\n\n- We need help writing tests and documentation!\n\n\n# License \n- Apache\n',
    'author': 'Zeta Team',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/zeta',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import torch
import torch.nn as nn
import torch.nn.functional as F


# QFTSPEmbedding
class QFTSPEmbedding(nn.Module):
    """
    QFTSPEmbedding with multiple collapse mechanisms.

    This module allows for different ways of collapsing the superposition of embeddings,
    based on the provided context and selected mechanism.
    """

    def __init__(
        self,
        vocab_size: int,
        dim: int,
        collapse_mode: str = "weighted_sum",
        **kwargs,
    ):
        super(QFTSPEmbedding, self).__init__()
        self.dim = dim
        self.collapse_mode = collapse_mode
        self.base_embeddings = nn.Embedding(vocab_size, dim)
        self.superposed_embeddings = nn.Embedding(vocab_size, dim)
        self.linear_transform = nn.Linear(2 * dim, dim)

    def forward(
        self, x: torch.Tensor, context_vector: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass of the QFTSPEmbedding module.

        Args:
            x (_type_): _description_
            context_vector (_type_): _description_
            collapse_mode (str, optional): _description_. Defaults to "weighted_sum".

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        base_embeds = self.base_embeddings(x)
        superposed_embeds = self.superposed_embeddings(x)

        if self.collapse_mode == "weighted_sum":
            collapsed_embeds = (
                base_embeds + context_vector.unsqueeze(-1) * superposed_embeds
            )
        elif self.collapse_mode == "dot_product":
            scale = torch.sum(
                superposed_embeds * context_vector.unsqueeze(-1),
                dim=-1,
                keepdim=True,
            )
            collapsed_embeds = base_embeds + scale * superposed_embeds
        elif self.collapse_mode == "cosine_similarity":
            scale = F.cosine_similarity(
                superposed_embeds, context_vector.unsqueeze(-1), dim=-1
            ).unsqueeze(-1)
            collapsed_embeds = base_embeds + scale * superposed_embeds
        elif self.collapse_mode == "gated":
            gate = torch.sigmoid(context_vector)
            collapsed_embeds = (
                base_embeds + gate.unsqueeze(-1) * superposed_embeds
            )
        elif self.collapse_mode == "concat_linear":
            concatenated = torch.cat([base_embeds, superposed_embeds], dim=-1)
            collapsed_embeds = self.linear_transform(concatenated)
        else:
            raise ValueError("Invalid collapse mode selected")

        return collapsed_embeds


# # Example Usage
# vocab_size = 10000
# dim = 512

# model = QFTSPEmbedding(vocab_size, dim)
# x = torch.randint(0, vocab_size, (1, 10))
# context_vector = torch.rand(1, 10)

# # Test different collapse modes
# for mode in ['weighted_sum', 'dot_product', 'cosine_similarity', 'gated', 'concat_linear']:
#     embeddings = model(x, context_vector, collapse_mode=mode)
#     print(f"Collapse mode: {mode}, Embeddings shape: {embeddings.shape}")



import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn


class MLPBlock(nn.Module):
    """MLPBlock

    Args:
        dim (int): [description]
    """

    def __init__(self, dim: int, hidden_dim: int):
        super(MLPBlock, self).__init__()
        self.dim = dim
        self.hidden_dim = hidden_dim
        self.dense1 = nn.Linear(dim, hidden_dim)
        self.dense2 = nn.Linear(hidden_dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of MLPBlock

        Args:
            x (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        y = self.dense1(x)
        y = F.gelu(y)
        return self.dense2(y)


class MixerBlock(nn.Module):
    """MixerBlock


    Args:
        mlp_dim (int): [description]
        channels_dim (int): [description]
    """

    def __init__(self, mlp_dim: int, channels_dim: int):
        super(MixerBlock, self).__init__()
        self.norm1 = nn.LayerNorm(channels_dim)
        self.tokens_mlp = MLPBlock(mlp_dim, mlp_dim)

        self.norm2 = nn.LayerNorm(channels_dim)
        self.channel_mlp = MLPBlock(mlp_dim, mlp_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of MixerBlock

        Args:
            x (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        y = self.norm1(x)
        y = rearrange(y, "n c t -> n t c")
        y = self.tokens_mlp(y)
        y = rearrange(y, "n t c -> n c t")
        x = x + y
        y = self.norm2(x)
        return x + self.channel_mlp(y)


class MLPMixer(nn.Module):
    """MLPMixer

    Args:
        num_classes (int): [description]
        num_blocks (int): [description]
        patch_size (int): [description]
        hidden_dim (int): [description]
        tokens_mlp_dim (int): [description]
        channels_mlp_dim (int): [description]

    Examples:
        >>> from zeta.nn import MLPMixer
        >>> model = MLPMixer(10, 8, 16, 32, 64, 64)
        >>> x = torch.randn(32, 3, 224, 224)
        >>> model(x).shape
        torch.Size([32, 10])


    """

    def __init__(
        self,
        num_classes: int,
        num_blocks: int,
        patch_size: int,
        hidden_dim: int,
        tokens_mlp_dim: int,
        channels_mlp_dim: int,
    ):
        super(MLPMixer, self).__init__()
        self.stem = nn.Conv2d(
            hidden_dim, hidden_dim, kernel_size=patch_size, stride=patch_size
        )
        self.mixer_blocks = nn.ModuleList(
            [
                MixerBlock(tokens_mlp_dim, channels_mlp_dim)
                for _ in range(num_blocks)
            ]
        )
        self.pred_head_layernorm = nn.LayerNorm(hidden_dim)
        self.head = nn.Linear(hidden_dim, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of MLPMixer

        Args:
            x (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        x = self.stem(x)
        x = rearrange(x, "n c h w -> n (h w) c")
        for mixer_block in self.mixer_blocks:
            x = mixer_block(x)
        x = self.pred_head_layernorm(x)
        x = x.mean(dim=1)
        return self.head(x)


# # Example of creating a model instance
# mlp_mixer = MLPMixer(
#     num_classes=10,
#     num_blocks=8,
#     patch_size=16,
#     hidden_dim=512,
#     tokens_mlp_dim=512,
#     channels_mlp_dim=512,
# )

# # Example input tensor
# example_input = torch.randn(
#     1, 512, 32, 32
# )  # Batch size of 1, 512 channels, 32x32 image
# output = mlp_mixer(example_input)
# print(
#     output.shape
# )  # Should output the shape corresponding to the number of classes



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

import math

import torch
from torch import nn


class RelativePositionBias(nn.Module):
    def __init__(
        self,
        bidirectional: int = True,
        num_buckets: int = 32,
        max_distance: int = 128,
        num_heads: int = 1,
    ):
        super().__init__()
        self.bidirectional = bidirectional
        self.num_buckets = num_buckets
        self.max_distance = max_distance
        self.num_heads = num_heads
        self.relative_attention_bias = nn.Embedding(
            self.num_buckets, self.num_heads
        )

    @staticmethod
    def _relative_position_bucket(
        relative_position, bidirectional=True, num_buckets=32, max_distance=128
    ):
        ret = 0
        n = -relative_position
        if bidirectional:
            num_buckets //= 2
            ret += (n < 0).to(torch.long) * num_buckets
            n = torch.abs(n)
        else:
            n = torch.max(n, torch.zeros_like(n))

        max_exact = num_buckets // 2
        is_small = n < max_exact

        val_if_large = max_exact + (
            torch.log(n.float() / max_exact)
            / math.log(max_distance / max_exact)
            * (num_buckets - max_exact)
        ).to(torch.long)
        val_if_large = torch.min(
            val_if_large, torch.full_like(val_if_large, num_buckets - 1)
        )

        ret += torch.where(is_small, n, val_if_large)
        return ret

    def compute_bias(self, qlen, klen, step=None):
        step = 0 if step is None else step
        context_position = torch.arange(
            step,
            step + qlen,
            dtype=torch.long,
            device=self.relative_attention_bias.weight.device,
        )[:, None]
        memory_position = torch.arange(
            klen,
            dtype=torch.long,
            device=self.relative_attention_bias.weight.device,
        )[None, :]
        relative_position = (
            memory_position - context_position
        )  # shape (qlen, klen)

        rp_bucket = self._relative_position_bucket(
            relative_position,  # shape (qlen, klen)
            bidirectional=self.bidirectional,
            num_buckets=self.num_buckets,
            max_distance=self.max_distance,
        )
        rp_bucket = rp_bucket.to(self.relative_attention_bias.weight.device)
        values = self.relative_attention_bias(
            rp_bucket
        )  # shape (qlen, klen, num_heads)
        values = values.permute([2, 0, 1]).unsqueeze(
            0
        )  # shape (1, num_heads, qlen, klen)
        return values

    def forward(self, batch_size, qlen, klen, step=None):
        # shape (batch * num_heads, qlen, klen)
        return (
            self.compute_bias(qlen, klen, step)
            .repeat(batch_size, 1, 1, 1)
            .view(-1, qlen, klen)
        )
